@@ -18,7 +18,7 @@ use Carp qw(croak);
 @ISA     = qw(Business::OnlinePayment::HTTPS);
 $me      = 'Business::OnlinePayment::Litle';
 $DEBUG   = 0;
-$VERSION = '0.910';
+$VERSION = '0.912';
 
 =head1 NAME
 
@@ -101,7 +101,9 @@ The following actions are valid
 
 =head2 Fields
 
-Mostdata fields nto part of the BOP standard can be added to the content hash directly, and will be used
+Most data fields not part of the BOP standard can be added to the content hash directly, and will be used
+
+Most data fields will truncate extra characters to conform to the Litle XML length requirements.  Some fields (mostly amount fields) will error if your data exceeds the allowed length.
 
 =head2 Products
 
@@ -112,10 +114,10 @@ Part of the enhanced data for level III Interchange rates
         sku         =>  'sku',
         quantity    =>  1,
         units       =>  'Months'
-        amount      =>  500,  ## currently I don't reformat this, $5.00
+        amount      =>  '5.00',
         discount    =>  0,
         code        =>  1,
-        cost        =>  500,
+        cost        =>  '5.00',
     },
     {   description =>  'Second Product',
         sku         =>  'sku',
@@ -124,7 +126,7 @@ Part of the enhanced data for level III Interchange rates
         amount      =>  1500,
         discount    =>  0,
         code        =>  2,
-        cost        =>  500,
+        cost        =>  '5.00',
     }
 
     ],
@@ -133,7 +135,7 @@ Part of the enhanced data for level III Interchange rates
 
 =head1 SPECS
 
-Currently uses the Litle XML specifications version 7.2
+Currently uses the Litle XML specifications version 8.12
 
 =head1 TESTING
 
@@ -231,8 +233,6 @@ sub map_fields {
     );
     $content->{'TransactionType'} = $actions{$action} || $action;
 
-    $content->{'company_phone'} =~ s/\D//g if $content->{'company_phone'};
-
     my $type_translate = {
         'VISA card'                   => 'VI',
         'MasterCard'                  => 'MC',
@@ -260,8 +260,6 @@ sub map_fields {
       ? 'Existing'
       : 'New';    # new/Existing
 
-    $content->{'expiration'} =~ s/\D+//g;
-
     $content->{'deliverytype'} = 'SVC';
 
     # stuff it back into %content
@@ -271,9 +269,6 @@ sub map_fields {
             $_->{'itemSequenceNumber'} = $count++;
         }
     }
-    if ( $content->{'cvv2'} && length( $content->{'cvv2'} ) > 4 ) {
-        croak "CVV2 has too many characters";
-    }
 
     if( $content->{'velocity_check'} && (
         $content->{'velocity_check'} != 0 
@@ -295,6 +290,86 @@ sub map_fields {
     return $content;
 }
 
+=head2 format_misc_field
+
+Used internally to guarentee that XML data will conform to the Litle spec.
+  maxLen - The maximum length allowed (extra bytes will be truncated)
+  minLen - The minimum length allowed
+  errorOnLength - boolean
+    0 - truncate any extra bytes
+    1 - error if the length is out of bounds
+  isRequired - boolean
+    0 - ignore undefined values
+    1 - error if the value is not defined
+
+$tx->format_misc_field( \%content, [field, maxLen, minLen, errorOnLength, isRequired] );
+$tx->format_misc_field( \%content, ['amount',   0,     12,             0,          0] );
+
+=cut
+
+sub format_misc_field {
+    my ($self, $content, $trunc) = @_;
+
+    use bytes; # make sure we truncate on bytes, not characters
+
+    if( defined $content->{ $trunc->[0] } ) {
+      my $len = bytes::length( $content->{ $trunc->[0] } );
+      if ( $trunc->[3] && $trunc->[2] && $len != 0 && $len < $trunc->[2] ) {
+        # Zero is a valid length (mostly for cvv2 value)
+        croak "$trunc->[0] has too few characters";
+      }
+      elsif ( $trunc->[3] && $trunc->[1] && $len > $trunc->[1] ) {
+        croak "$trunc->[0] has too many characters";
+      }
+      $content->{ $trunc->[0] } = bytes::substr($content->{ $trunc->[0] } , 0, $trunc->[1] );
+      #warn "$trunc->[0] => $len => $content->{ $trunc->[0] }\n" if $DEBUG;
+    }
+    elsif ( $trunc->[4] ) {
+      croak "$trunc->[0] is required";
+    }
+}
+
+=head2 format_amount_field
+
+Used internally to change amounts from the BOP "5.00" format to the format expected by Litle "500"
+
+$tx->format_amount_field( \%content, 'amount' );
+
+=cut
+
+sub format_amount_field {
+    my ($self, $data, $field) = @_;
+    if (defined ( $data->{$field} ) ) {
+        $data->{$field} = sprintf( "%.2f", $data->{$field} );
+        $data->{$field} =~ s/\.//g;
+    }
+}
+
+=head2 format_phone_field
+
+Used internally to strip invalid characters from phone numbers. IE "1 (800).TRY-THIS" becomes "18008788447"
+
+$tx->format_phone_field( \%content, 'company_phone' );
+
+=cut
+
+sub format_phone_field {
+    my ($self, $data, $field) = @_;
+    if (defined ( $data->{$field} ) ) {
+        my $convertPhone = {
+            'a' => 2, 'b' => 2, 'c' => 2,
+            'd' => 3, 'e' => 3, 'f' => 3,
+            'g' => 4, 'h' => 4, 'i' => 4,
+            'j' => 5, 'k' => 5, 'l' => 5,
+            'm' => 6, 'n' => 6, 'o' => 6,
+            'p' => 7, 'q' => 7, 'r' => 7, 's' => 7,
+            't' => 8, 'u' => 8, 'v' => 8,
+            'w' => 9, 'x' => 9, 'y' => 9, 'z' => 9,
+        };
+        $data->{$field} =~ s/(\D)/$$convertPhone{lc($1)}||''/eg;
+    }
+}
+
 sub map_request {
     my ( $self, $content ) = @_;
 
@@ -307,27 +382,69 @@ sub map_request {
     $self->required_fields(@required_fields);
 
     # for tabbing
-    # clean up the amount to the required format
-    my $amount;
-    if ( defined( $content->{amount} ) ) {
-        $amount = sprintf( "%.2f", $content->{amount} );
-        $amount =~ s/\.//g;
+    # set dollar amounts to the required format (eg $5.00 should be 500)
+    foreach my $field ( 'amount', 'salesTax', 'discountAmount', 'shippingAmount', 'dutyAmount' ) {
+        $self->format_amount_field($content, $field);
     }
     
+    # make sure the date is in MMYY format
+    $content->{'expiration'} =~ s/^(\d{1,2})\D*\d*?(\d{2})$/$1$2/;
+
+    if ( ! defined $content->{'description'} ) { $content->{'description'} = ''; } # shema req
+    $content->{'description'} =~ s/[^\w\s\*\,\-\'\#\&\.]//g;
+
+    # only numbers are allowed in company_phone
+    $self->format_phone_field($content, 'company_phone');
+
     #  put in a list of constraints
     my @validate = (
-      [ 'city', 35 ],
-      [ 'address', 35 ],
-      [ 'state', 30 ],
-      [ 'name', 100 ],
+      # field,     maxLen, minLen, errorOnLength, isRequired
+      [ 'name',       100,      0,             0, 0 ],
+      [ 'email',      100,      0,             0, 0 ],
+      [ 'address',     35,      0,             0, 0 ],
+      [ 'city',        35,      0,             0, 0 ],
+      [ 'state',       30,      0,             0, 0 ], # 30 is allowed, but it should be the 2 char code
+      [ 'zip',         20,      0,             0, 0 ],
+      [ 'country',      3,      0,             0, 0 ], # should use iso 3166-1 2 char code
+      [ 'phone',       20,      0,             0, 0 ],
+
+      [ 'ship_name',  100,      0,             0, 0 ],
+      [ 'ship_email', 100,      0,             0, 0 ],
+      [ 'ship_address',35,      0,             0, 0 ],
+      [ 'ship_city',   35,      0,             0, 0 ],
+      [ 'ship_state',  30,      0,             0, 0 ], # 30 is allowed, but it should be the 2 char code
+      [ 'ship_zip',    20,      0,             0, 0 ],
+      [ 'ship_country', 3,      0,             0, 0 ], # should use iso 3166-1 2 char code
+      [ 'ship_phone',  20,      0,             0, 0 ],
+
+      #[ 'customerType',13,      0,             0, 0 ],
+
+      ['company_phone',13,      0,             0, 0 ],
+      [ 'description', 25,      0,             0, 0 ],
+
+      [ 'po_number',   17,      0,             0, 0 ],
+      [ 'salestax',     8,      0,             1, 0 ],
+      [ 'discount',     8,      0,             1, 0 ],
+      [ 'shipping',     8,      0,             1, 0 ],
+      [ 'duty',         8,      0,             1, 0 ],
+      ['invoice_number',15,     0,             0, 0 ], # TODO orderID = 25, invoiceReferenceNumber = 15
+      [ 'orderdate',   10,      0,             0, 0 ], # YYYY-MM-DD
+
+      [ 'card_type',    2,      2,             1, 0 ],
+      [ 'card_number', 25,     13,             1, 0 ],
+      [ 'expiration',   4,      4,             1, 0 ], # MMYY
+      [ 'cvv2',         4,      3,             1, 0 ],
+      # 'token' does not have a documented limit
+
+      [ 'customer_id', 25,      0,             0, 0 ],
     );
     foreach my $trunc ( @validate ) {
-      if( defined $content->{ $trunc->[0] } ) {
-        $content->{ $trunc->[0] } = substr($content->{ $trunc->[0] } , 0, $trunc->[1] );
-      }
-    };
+      $self->format_misc_field($content,$trunc);
+      #warn "$trunc->[0] => ".($content->{ $trunc->[0] }||'')."\n" if $DEBUG;
+    }
 
     tie my %billToAddress, 'Tie::IxHash', $self->revmap_fields(
+        content      => $content,
         name         => 'name',
         email        => 'email',
         addressLine1 => 'address',
@@ -340,6 +457,7 @@ sub map_request {
     );
 
     tie my %shipToAddress, 'Tie::IxHash', $self->revmap_fields(
+        content      => $content,
         name         => 'ship_name',
         email        => 'ship_email',
         addressLine1 => 'ship_address',
@@ -352,30 +470,45 @@ sub map_request {
     );
 
     tie my %customerinfo, 'Tie::IxHash',
-      $self->revmap_fields( customerType => 'customerType', );
-
-    my $description =
-      $content->{'description'}
-      ? substr( $content->{'description'}, 0, 25 )
-      : '';    # schema req
-    $description =~ s/[^\w\s\*\,\-\'\#\&\.]//g;
+      $self->revmap_fields(
+        content      => $content,
+        customerType => 'customerType',
+      );
 
     tie my %custombilling, 'Tie::IxHash',
       $self->revmap_fields(
+        content      => $content,
         phone      => 'company_phone',
-        descriptor => \$description,
+        descriptor => 'description',
       );
 
     ## loop through product list and generate linItemData for each
     #
     my @products = ();
-    if( scalar( @{ $content->{'products'} } ) < 100 ){
-      foreach my $prod ( @{ $content->{'products'} } ) {
-          $prod->{'description'} = substr( $prod->{'description'}, 0, 25 );
-          $prod->{'code'} = substr( $prod->{'code'}, 0, 12 );
+    if( defined $content->{'products'} && scalar( @{ $content->{'products'} } ) < 100 ){
+      foreach my $prodOrig ( @{ $content->{'products'} } ) {
+          # use a local copy of prod so that we do not have issues if they try to submit more then once.
+          my %prod = %$prodOrig;
+          foreach my $field ( 'tax','amount','totalwithtax','discount' ) {
+            # Note: DO NOT format 'cost', it uses the decimal format
+            $self->format_amount_field(\%prod, $field);
+          }
+
+          my @validate = (
+            # field,     maxLen, minLen, errorOnLength, isRequired
+            [ 'description', 26,      0,             0, 0 ],
+            [ 'tax',          8,      0,             1, 0 ],
+            [ 'amount',       8,      0,             1, 0 ],
+            [ 'totalwithtax', 8,      0,             1, 0 ],
+            [ 'discount',     8,      0,             1, 0 ],
+            [ 'code',        12,      0,             0, 0 ],
+            [ 'cost',        12,      0,             1, 0 ],
+          );
+          foreach my $trunc ( @validate ) { $self->format_misc_field(\%prod,$trunc); }
+
           tie my %lineitem, 'Tie::IxHash',
             $self->revmap_fields(
-              content              => $prod,
+              content              => \%prod,
               itemSequenceNumber   => 'itemSequenceNumber',
               itemDescription      => 'description',
               productCode          => 'code',
@@ -386,7 +519,7 @@ sub map_request {
               lineItemTotalWithTax => 'totalwithtax',
               itemDiscountAmount   => 'discount',
               commodityCode        => 'code',
-              unitCost             => 'cost',
+              unitCost             => 'cost', # This "amount" field uses decimals
             );
           push @products, \%lineitem;
       }
@@ -395,6 +528,7 @@ sub map_request {
     #
     #
     tie my %enhanceddata, 'Tie::IxHash', $self->revmap_fields(
+        content                => $content,
         customerReference      => 'po_number',
         salesTax               => 'salestax',
         discountAmount         => 'discount',
@@ -406,25 +540,28 @@ sub map_request {
     );
 
     tie my %card, 'Tie::IxHash', $self->revmap_fields(
+        content            => $content,
         type               => 'card_type',
         number             => 'card_number',
         expDate            => 'expiration',
         cardValidationNum  => 'cvv2',
-        cardAuthentication => '3ds',          # is this what we want to name it?
     );
 
     tie my %token, 'Tie::IxHash', $self->revmap_fields(
+        content            => $content,
         litleToken         => 'token',
         expDate            => 'expiration',
         cardValidationNum  => 'cvv2',
     );
 
     tie my %processing, 'Tie::IxHash', $self->revmap_fields(
+        content               => $content,
         bypassVelocityCheck   => 'velocity_check',
     );
 
     tie my %cardholderauth, 'Tie::IxHash',
       $self->revmap_fields(
+        content                     => $content,
         authenticationValue         => '3ds',
         authenticationTransactionId => 'visaverified',
         customerIpAddress           => 'ip',
@@ -435,8 +572,9 @@ sub map_request {
 
     if ( $action eq 'sale' ) {
         tie %req, 'Tie::IxHash', $self->revmap_fields(
+            content       => $content,
             orderId       => 'invoice_number',
-            amount        => \$amount,
+            amount        => 'amount',
             orderSource   => 'orderSource',
             billToAddress => \%billToAddress,
             card          => \%card,
@@ -451,12 +589,14 @@ sub map_request {
     }
     elsif ( $action eq 'authorization' ) {
         tie %req, 'Tie::IxHash', $self->revmap_fields(
+            content       => $content,
             orderId       => 'invoice_number',
-            amount        => \$amount,
+            amount        => 'amount',
             orderSource   => 'orderSource',
             billToAddress => \%billToAddress,
             card          => \%card,
             token         => $content->{'token'} ? \%token : {},
+
             #cardholderAuthentication    =>  \%cardholderauth,
             processingInstructions  =>  \%processing,
             customBilling => \%custombilling,
@@ -467,8 +607,9 @@ sub map_request {
         push @required_fields, qw( order_number amount );
         tie %req, 'Tie::IxHash',
           $self->revmap_fields(
+            content      => $content,
             litleTxnId   => 'order_number',
-            amount       => \$amount,
+            amount       => 'amount',
             enhancedData => \%enhanceddata,
             processingInstructions  =>  \%processing,
           );
@@ -479,8 +620,9 @@ sub map_request {
        if( $content->{'order_number'} ){
           push @required_fields, qw( order_number amount );
           tie %req, 'Tie::IxHash', $self->revmap_fields(
+              content       => $content,
               litleTxnId    => 'order_number',
-              amount        => \$amount,
+              amount        => 'amount',
               customBilling => \%custombilling,
               processingInstructions  =>  \%processing,
           );
@@ -489,8 +631,9 @@ sub map_request {
        else {
           push @required_fields, qw( invoice_number amount );
           tie %req, 'Tie::IxHash', $self->revmap_fields(
+              content       => $content,
               orderId       => 'invoice_number',
-              amount        => \$amount,
+              amount        => 'amount',
               orderSource   => 'orderSource',
               billToAddress => \%billToAddress,
               card          => \%card,
@@ -504,6 +647,7 @@ sub map_request {
         push @required_fields, qw( order_number );
         tie %req, 'Tie::IxHash',
           $self->revmap_fields( 
+            content                 => $content,
             litleTxnId              => 'order_number', 
             processingInstructions  =>  \%processing,
           );
@@ -512,14 +656,16 @@ sub map_request {
         push @required_fields, qw( order_number amount );
         tie %req, 'Tie::IxHash',
           $self->revmap_fields(
+            content    => $content,
             litleTxnId => 'order_number',
-            amount     => \$amount,
+            amount     => 'amount',
           );
     }
     elsif ( $action eq 'accountUpdate' ) {
         push @required_fields, qw( card_number expiration );
         tie %req, 'Tie::IxHash',
           $self->revmap_fields(
+            content => $content,
             orderId => 'customer_id',
             card    => \%card,
           );
@@ -538,7 +684,9 @@ sub submit {
     $self->is_success(0);
 
     my %content = $self->content();
+    warn 'Pre processing: '.Dumper(\%content) if $DEBUG;
     my $req     = $self->map_request( \%content );
+    warn 'Post processing: '.Dumper(\%content) if $DEBUG;
     my $post_data;
 
     my $writer = new XML::Writer(
@@ -551,6 +699,7 @@ sub submit {
     ## set the authentication data
     tie my %authentication, 'Tie::IxHash',
       $self->revmap_fields(
+        content  => \%content,
         user     => 'login',
         password => 'password',
       );
@@ -589,16 +738,18 @@ sub submit {
     $writer->end();
     ## END XML Generation
 
-    my ( $page, $server_response, %headers ) = $self->https_post($post_data);
     $self->{'_post_data'} = $post_data;
     warn $self->{'_post_data'} if $DEBUG;
+    my ( $page, $server_response, %headers ) = $self->https_post($post_data);
 
     warn Dumper $page, $server_response, \%headers if $DEBUG;
 
     my $response = {};
     if ( $server_response =~ /^200/ ) {
-        $response = XMLin($page);
-        if ( exists( $response->{'response'} ) && $response->{'response'} == 1 )
+        if ( ! eval { $response = XMLin($page); } ) {
+            die "XML PARSING FAILURE: $@";
+        }
+        elsif ( exists( $response->{'response'} ) && $response->{'response'} == 1 )
         {
             ## parse error type error
             warn Dumper( $response, $self->{'_post_data'} );
@@ -612,6 +763,8 @@ sub submit {
         }
     }
     else {
+        $server_response =~ s/[\r\n\s]+$//; # remove newline so you can see the error in a linux console
+        if ( $server_response =~ /^900/ ) { $server_response .= ' - verify Litle has whitelisted your IP'; }
         die "CONNECTION FAILURE: $server_response";
     }
     $self->{_response} = $response;
@@ -688,7 +841,7 @@ sub parse_batch_response {
     };
 }
 
-=head1 add_item
+=head2 add_item
 
 A new method, not supported under BOP yet, but interface to adding multiple entries, so we can write and interface with batches
 
@@ -809,8 +962,10 @@ sub create_batch {
 
         my $response = {};
         if ( $server_response =~ /^200/ ) {
-            $response = XMLin($page);
-            if ( exists( $response->{'response'} )
+            if ( ! eval { $response = XMLin($page); } ) {
+                die "XML PARSING FAILURE: $@";
+            }
+            elsif ( exists( $response->{'response'} )
                 && $response->{'response'} == 1 )
             {
                 ## parse error type error
@@ -912,8 +1067,10 @@ sub send_rfr {
 
     my $response = {};
     if ( $server_response =~ /^200/ ) {
-        $response = XMLin($page);
-        if ( exists( $response->{'response'} ) && $response->{'response'} == 1 )
+        if ( ! eval { $response = XMLin($page); } ) {
+            die "XML PARSING FAILURE: $@";
+        }
+        elsif ( exists( $response->{'response'} ) && $response->{'response'} == 1 )
         {
             ## parse error type error
             warn Dumper( $response, $self->{'_post_data'} );
@@ -963,20 +1120,19 @@ sub retrieve_batch {
 
         $sftp->setcwd("outbound")
           or die "Cannot change working directory ", $sftp->error;
-        ## save the file out, can't put directly from var, and is multibyte, so issues from filehandle
-        my $io = IO::String->new($post_data);
-        tie *IO, 'IO::String';
 
         my $filename = $opts{'batch_id'};
-        $sftp->get( "$filename.asc", $io )
+        $post_data = $sftp->get_content( $filename )
           or die "Cannot GET $filename", $sftp->error;
         $self->is_success(1);
         warn $post_data if $DEBUG;
     }
 
     my $response = {};
-    $response = XMLin($post_data);
-    if ( exists( $response->{'response'} ) && $response->{'response'} == 1 ) {
+    if ( ! eval { $response = XMLin($post_data); } ) {
+        die "XML PARSING FAILURE: $@";
+    }
+    elsif ( exists( $response->{'response'} ) && $response->{'response'} == 1 ) {
         ## parse error type error
         warn Dumper( $response, $self->{'_post_data'} );
         $self->error_message( $response->{'message'} );
@@ -1018,6 +1174,7 @@ sub revmap_fields {
         %content = %{ delete( $map{'content'} ) };
     }
     else {
+        warn "WARNING: This content has no been pre-processed with map_fields";
         %content = $self->content();
     }
 
@@ -1063,6 +1220,7 @@ sub _xmlwrite {
     }
     else {
         $writer->startTag($item);
+	utf8::decode($value); # prevent double byte corruption in the xml output
         $writer->characters($value);
         $writer->endTag($item);
     }
